# SPDX-License-Identifier: MIT
#
# TFTP Proxy
#
# This script is TFTP proxy that works with PowerShell7.
# Ana also it supports RRQ and WRQ as specified in RFC1350.
#
# (C) 2025 Yutaka Hirata(YOULAB)
# https://yutakasan.11ya.net/
# https://crowdworks.jp/public/employees/2224566
#

<#
.SYNOPSIS

TFTP Proxy

.DESCRIPTION

This script is TFTP proxy that works with PowerShell7.
Ana also it supports RRQ and WRQ as specified in RFC1350.

.EXAMPLE

.\udp_proxy.ps1 -TFTPServerIPAddress 192.168.1.1

.EXAMPLE

.\udp_proxy.ps1 -TFTPServerIPAddress 192.168.1.1 -Progress

.EXAMPLE

.\udp_proxy.ps1 -ListenIPAddress 10.0.0.1 -ListenPort 69 -TFTPServerIPAddress 192.168.1.1 -TFTPServerPort 69 -TimeoutValue 5 -LogVerbose -Progress


.NOTES
    Author: Yutaka Hirata(YOULAB)
    Web: https://yutakasan.11ya.net/
    Version: 1.0
    Last Updated: January 3, 2025
#>

# Command line parameters
Param (
    # Listen IP address 
    [string]
    $ListenIPAddress = "0.0.0.0",    

    # Listen Port number 
    [int]
    $ListenPort = 69,                  
    
    # TFTP server IP address (Mandatory)
    [Parameter(Mandatory=$true)]
    [string]
    $TFTPServerIPAddress,  

    # TFTP server Port number
    [int]
    $TFTPServerPort = 69,             

    # Session timeout (seconds)
    [int]
    $TimeoutValue = 15,

    # Verbose mode for Debug
    [switch] 
    $LogVerbose = $false,

    # Show a progress however script will slow down.
    [switch] 
    $Progress = $false
)


# Strict grammar checks
#
# NOTE: Must be written after above Param definition
# because the Param will become an error.
Set-StrictMode -Version Latest

# Command line parameters
$global:client_IP = $null
$global:client_Port = $null
$global:server_IP = $null
$global:server_Port = $null
$global:logverbose = $null
$global:progress = $null
$global:sessionTimeoutValue = $null


# Information for tftp client
$global:clientEndPoint = $null
$global:udpClient = $null
$global:udpClientSocket = $null

# Information for tftp server
$global:serverEndPoint = $null
$global:udpServer = $null
$global:udpServerSocket = $null

enum OpcodeType {
    RRQ
    WRQ
    INVALID
}
# TFTP Operation Code
$global:opcode = [OpcodeType]::INVALID


# Retrieve the command line parameters and setup to global variables.
Function CheckCommandLineParameters() {
    $global:client_IP = $ListenIPAddress
    $global:client_Port = $ListenPort
    $global:server_IP = $TFTPServerIPAddress
    $global:server_Port = $TFTPServerPort
    $global:logverbose = $LogVerbose
    $global:progress = $Progress
    $global:sessionTimeoutValue = $TimeoutValue

    # Show all options.
    $info = @"
Arguments:
    Listen IP Address           : $global:client_IP
    Listen Port                 : $global:client_Port
    TFTP Server IP Address      : $global:server_IP
    TFTP Server Port            : $global:server_Port
    Verbose Mode                : $global:logverbose
    Progress Display            : $global:progress
    Session Timeout Value(sec)  : $global:sessionTimeoutValue
"@

    Write-Host $info -ForegroundColor DarkYellow
}

# Shows details of the stack trace when an exception occurs.
function Resolve-Error($ErrorRecord=$Error[0])
{
   $ErrorRecord | Format-List * -Force
   $ErrorRecord.InvocationInfo |Format-List *
   $Exception = $ErrorRecord.Exception
   for ($i = 0; $Exception; $i++, ($Exception = $Exception.InnerException))
   {   "$i" * 80
       $Exception |Format-List * -Force
   }
}

# Show hexadecimal dump.
function HexDump {
    param (
        [Parameter(Mandatory, Position = 0)]
        [ValidateNotNullOrEmpty()]
        [byte[]]$Data
    )

    $lineLength = 16  # Number of bytes per line

    # Process the bytes in chunks of $lineLength
    for ($i = 0; $i -lt $Data.Length; $i += $lineLength) {
        $chunk = $Data[$i..([math]::Min($i + $lineLength - 1, $Data.Length - 1))]

        # Offset
        $offset = "{0:X8}" -f $i

        # Hexadecimal representation
        $hex = ($chunk | ForEach-Object { "{0:X2}" -f $_ }) -join ' '

        # ASCII representation (printable characters only)
        $ascii = ($chunk | ForEach-Object {
            if ($_ -ge 32 -and $_ -le 126) {
                [char]$_
            } else {
                '.'
            }
        }) -join ''

        # Print the line
        write-host("{0}  {1,-48}  {2}" -f $offset, $hex, $ascii)
    }
}

# Send packets to TFTP client.
Function sendDataToClient($data) {
    $ret = $true

    $dataLen = $data.Length
    $sentByte = $global:udpClient.Send($data, $dataLen, $global:clientEndPoint)
    if ($global:logverbose) {
        write-host "Send to client $dataLen($sentByte) bytes"
    }
    if ($sentByte -ne $dataLen) {
        write-host "Failed to send to client ${$global:clientIP}:${$global:clientPort}" -ForegroundColor Red
        $ret = $false
    }    

    return $ret
}

# Receive packets from TFTP client.
Function recvDataFromClient() {
    # If this script is blocking in a .NET library function, 
    # CTRL+C will not work. Therefore, it is handled by polling.
    while ($true) {
        if ($global:udpClientSocket.Poll(0, [System.Net.Sockets.SelectMode]::SelectRead)) {
            $data = $global:udpClient.Receive([ref]$global:clientEndPoint)
            break
        }
        start-sleep -Milliseconds 1
    }

    if ($global:logverbose) {
        $clientIP = $global:clientEndPoint.Address.ToString()
        $clientPort = $global:clientEndPoint.Port
        write-host "Receive from client ${clientIP}:${clientPort}"
        HexDump $data
    }

    return $data
}

# Send packets to TFTP server.
Function sendDataToServer($data) {
    $ret = $true

    $dataLen = $data.Length
    $sentByte = $global:udpServer.Send($data, $dataLen, $global:serverEndPoint)
    if ($global:logverbose) {
        write-host "Send to server $dataLen($sentByte) bytes"
    }
    if ($sentByte -ne $dataLen) {
        write-host "Failed to send to server ${$global:server_IP}:${$global:server_Port}" -ForegroundColor Red
        $ret = $false
    }    

    return $ret
}

# Receive packets from TFTP server.
Function recvDataFromServer() {
    # If this script is blocking in a .NET library function, 
    # CTRL+C will not work. Therefore, it is handled by polling.
    while ($true) {
        if ($global:udpServerSocket.Poll(0, [System.Net.Sockets.SelectMode]::SelectRead)) {
            $data = $global:udpServer.Receive([ref]$global:serverEndPoint)
            break
        }
        start-sleep -Milliseconds 1
    }

    if ($global:logverbose) {
        $serverIP = $global:serverEndPoint.Address.ToString()
        $serverPort = $global:serverEndPoint.Port
        write-host "Receive from server ${serverIP}:${serverPort}"
        HexDump $data
    }

    return $data
}

# Initializes UDP client and server socket.
Function initialSocket() {
    # Making UDP client 
    $ipaddr = [System.Net.IPAddress]::Parse($global:client_IP)
    $port = $global:client_Port
    $global:clientEndPoint = [System.Net.IPEndPoint]::new($ipaddr, $port)
    $global:udpClient = [System.Net.Sockets.UdpClient]::new($global:clientEndPoint)
    $global:udpClientSocket = $global:udpClient.Client
    $global:udpClientSocket | out-null

    # Making UDP server
    $ipaddr = [Net.IPAddress]::Parse($global:server_IP)
    $port = $global:server_Port
    $global:serverEndPoint = [System.Net.IPEndPoint]::new($ipaddr, $port)
    $global:serverEndPoint | out-null
    $global:udpServer = [System.Net.Sockets.UdpClient]::new()
    $global:udpServerSocket = $global:udpServer.Client
    $global:udpServerSocket | out-null
}

# Initialize TFTP server endpoint information. 
# The first communication uses UDP/69. however after the second time, 
# a random port is used. So the first endpoint is overwritten.
Function reSetupEndPoint() {
    $ipaddr = [Net.IPAddress]::Parse($global:server_IP)
    $port = $global:server_Port
    $global:serverEndPoint = [System.Net.IPEndPoint]::new($ipaddr, $port)
    $global:serverEndPoint | out-null
}

# Snoop on an UDP packet to get TFTP OpCode.
Function checkOperationCode($data) {
    $str = "Invalid"

    $global:opcode = [OpcodeType]::INVALID
    $color = "Red"

    if ($data.Length -ge 2) {
        $opcode = ($data[0] * 0xFF + $data[1])

        switch($opcode) {
            1 { $str = "RRQ"
                $global:opcode = [OpcodeType]::RRQ
                $color = "Green"
            }

            2 { $str = "WRQ"
                $global:opcode = [OpcodeType]::WRQ
                $color = "DarkCyan"
            }
        } 
    } 

    Write-Host "TFTP operation is $str" -ForegroundColor $color
}

# Sends the TFTP client's initial communication 
# information to the TFTP server (UDP/69).
Function firstHandshakeProcess() {
    $ret = $true

    reSetupEndPoint

    $data = recvDataFromClient

    checkOperationCode $data

    $ret = sendDataToServer $data
    if (-not $ret) {
        $ret = $false
    }

    return $ret
}


$frames = @("|", "/", "-", "\")
$currentIndex = 0
function Write-NextFrame {
    param (
        [array]$framesRef,  
        [ref]$indexRef      
    )

    Write-Host -NoNewline "`b"
    Write-Host -NoNewline $framesRef[$indexRef.Value]
    $indexRef.Value = ($indexRef.Value + 1) % $framesRef.Length
}

# Starts TFTP proxy
Function startProxyServer() {   
    initialSocket

    try {
            while ($True) {
            Write-Host "TFTP Proxy started on port ${client_IP}:${client_Port}"

            $ret = firstHandshakeProcess
            if (-not $ret) {
                break
            }
            #Write-Host "First command was received from tftp client" -ForegroundColor Green

            $timeout = [TimeSpan]::FromMilliseconds($global:sessionTimeoutValue * 1000)

            while ($True) {
                if ($global:udpServerSocket.Poll($timeout, [System.Net.Sockets.SelectMode]::SelectRead)) {
                    $data = recvDataFromServer

                    $ret = sendDataToClient $data
                    if (-not $ret) {
                        break
                    }

                    if ($global:progress) {
                        #Write-NextFrame -framesRef $frames -indexRef ([ref]$currentIndex)
                        Write-Host -NoNewline "#"
                    }
                } else {
                    if ($global:opcode -eq [OpcodeType]::RRQ) {
                        if ($global:logverbose) {
                            Write-Host "Server timeout: May be finished tftp get" -ForegroundColor Green
                        }
                        break
                    }
                }

                if ($global:udpClientSocket.Poll($timeout, [System.Net.Sockets.SelectMode]::SelectRead)) {
                    $data = recvDataFromClient

                    $ret = sendDataToServer $data
                    if (-not $ret) {
                        break
                    }
                } else {
                    if ($global:opcode -eq [OpcodeType]::WRQ) {
                        if ($global:logverbose) {
                            Write-Host "Client timeout: May be finished tftp put" -ForegroundColor Green
                        }
                        break
                    }
                }
            }

        }
    } catch {
        Write-Host "An error occurred: $_" -ForegroundColor Red

        #Resolve-Error

    } finally {
        $global:udpClient.Close()
        $global:udpClientSocket.Dispose()

        $global:udpServer.Close()
        $global:udpServerSocket.Dispose()

        Write-Host "TFTP proxy stopped."
    }
}

#
# Main routine
#
CheckCommandLineParameters

while ($true) {
    startProxyServer
}
